#################################################################################
# Created       : Martin Foster
# Modified      : 20-Mar-2007 
#################################################################################
#
# Home - Home page/file handling module for Ethereal Realms 
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

package Ethereal::Home;						# Package definition

use LWP::UserAgent;						# World Wide Web agent
use strict;							# Strict handling

use Ethereal::Web;						# LWP user agent wrapper

#################################################################################
# Sub-Routines
#################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;					# Class reference
	my $database = shift;					# Database handle
	my $basedir  = shift;					# Base directory

	my $perm;						# Permissions set
	my $self;						# Self referencing

	#####################
	# Program files

	# Confirmations
	# Base directory
	$basedir = (defined($basedir))
	  ? $basedir
	  : $database->{'SYS'}{'SetBaseDir'};

	# Permissions set
	$perm    = (defined($database->{'SYS'}{'SetPerm'}))
	  ? $database->{'SYS'}{'SetPerm'}
	  : '0775';


	# Generate hash
	# For SEP use '/' for Unix and '\' for Windows.
	$self  = { 
		HANDLE  => $database,
		BASEDIR => $basedir,
		PERM    => $database->{'SYS'}{'SetPerm'},
		UMASK   => '0000',
		SEP     => '/'
	  };

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}


#####################
# Directory Clear 
#
# Allows for the removal of a users Home page.  This is of course assuming that
# there are no sub-directories, which the home page component does not allow.
# However if this is the case, the function is recursive.

sub DirClear
{
	#####################
	# Data members

	my $self     = shift;					# Reference to self
	my $dir      = shift;					# Directory name 

	my $file;						# File name
	my $sub;						# Subdirectory
	my $sep      = $self->{'SEP'};				# Seperator

	my @locales;						# Additional directories


	#####################
	# Program area

	# Open Directory Handle 
	# Litmus test
	if (-d $dir)
	{
	 	opendir(DIR, $dir) or die("Error: $dir cannot be accessed!");	

		# Read through directory listing
		while ($file = readdir(DIR))
		{
			# Skip parent and current 
			next if ($file eq '.');
			next if ($file eq '..');			

			# Remove folders recursively 
			# Add to list
			if (-d "$dir$sep$file") 
			{ push(@locales, "$dir$sep$file"); }

			# Remove individual file
			else
			{ unlink("$dir$sep$file"); }
		}	

		# Close handle
		closedir(DIR);

		# Call additional directories
		foreach $sub (@locales)
		{
			# Call subdirectory
			$self->DirClear("$sub");
		}

		# Remove directory
		return rmdir($dir);
	}

	else
	{
		# Simply send a warning
		warn("Warning: $dir does not exist!");
	}
}


#####################
# Dir Create
#
# Is used to standardize the way in which directories are created
# hiding any complexity with unasks.

sub DirCreate
{
	#####################
	# Data members

	my $self     = shift;					# Reference to self
	my $dir      = shift;					# Directory name

	my $perm     = $self->{'PERM'};				# Permissions
	my $umask    = $self->{'UMASK'};			# Masking


	#####################
	# Program area

	# Sets the mask
	umask(oct($umask));

	# Create directory
	# Return
	return (mkdir($dir, oct($perm)));	
}


#####################
# File Fetch
#
# Will retrieve a file using an ordinary every day web address
# as to allow for image sizing or manipulation to take place.

sub FileFetch
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $address  = shift;					# File location 

	my $basedir  = $self->{'BASEDIR'};			# Base working directory
	my $sep      = $self->{'SEP'};				# Seperator

	my $file;						# File name
	my $web;						# HTTP user agent


	#####################
	# Program area

	# Initiate hashing handle
	use Digest::MD5 qw(md5_hex);


	# Legacy Random
	srand();

	# Generate random name
	$file  =  'ethereal_' . md5_hex(rand());
	

	# Initialize handles
	$web = new Ethereal::Web($self->{'HANDLE'});

	# Retreive image
	# And save to file
	$web->Request($address, $basedir . $sep . $file);

	# Return filename
	return $file;
}


#####################
# File listing 
#
# Handles the recovery of file information from within a users personal home 
# folder.   This will return size and file count for convenience.

sub FileListing 
{
	#####################
	# Data members

	my $self     = shift;					# Reference to self
	my $dir      = shift;					# Directory name 
	my $files    = shift;					# Array of files
	my $home     = shift;					# Home hash

	my $basedir  = $self->{'BASEDIR'};			# Base working directory
	my $sep      = $self->{'SEP'};				# Seperator line

	my $file;						# File name
	my $count = 0;						# File count
	my $size  = 0;						# File size


	#####################
	# Program area

	# Open Directory Handle 
 	opendir(DIR, "$basedir$sep$dir") or return $size, $count;	

	# Read through directory listing
	while ($file = readdir(DIR))
	{
		# Possible skips
		next if ($file =~ /^\./);
		next if ($file eq $home->{'SetHEADER'});
		next if ($file eq $home->{'SetREADME'});			

		# Add to size and increment
		$size += -s "$basedir$sep$dir$sep$file";
		$count++;

		# Append to array
		push(@{$files}, $file); 
	}	

	# Close handle
	closedir(DIR);
	
	# Return values
	return $size, $count;
}



#####################
# FileUnlink
#
# As the name states, this will remove the file itself using
# native Perl methods.

sub FileUnlink
{
	#####################
	# Data members

	my $self     = shift;					# Reference to self
	my $file     = shift;					# Filename 

	my $basedir  = $self->{'BASEDIR'};			# Base working directory
	my $sep      = $self->{'SEP'};				# Seperator line


	#####################
	# Program area

	# Verify for existence
	if (-f "$basedir$sep$file")
	{
		# Remove file
		unlink("$basedir$sep$file");
	}
}


#####################
# File Write 
#
# As the name implies, this subroutine is designed to hide some of the more
# complex features of file writing.  This system will write whatever was
# received in the parameter list.   Overwriting existing data. 

sub FileWrite 
{
	#####################
	# Data members

	my $self     = shift;					# Reference to self
	my $file     = shift;					# Name of file
	my $contents = shift;					# Contents of file

	my $basedir  = $self->{'BASEDIR'};			# Base working directory
	my $sep      = $self->{'SEP'};				# Seperator line


	#####################
	# Program area

	# Setting umask
	umask(oct('0111'));

	# Open file
	if (open(FILE, ">$basedir$sep$file"))
	{
		# Write out contents
		print FILE $contents;

		# Close handle
		close(FILE);

		# Return true
		return 1;
	}

	# Return error
	return $!;
}


1;								# Return true
