################################################################################
# Created       : Martin Foster
# Modified      : 08-Apr-2007
################################################################################
#
# OpenID - Modules which will handle the complexities associated with the
#          OpenID authentication system.
# Copyright (C) 2000-2007  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 4-3526 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

package Ethereal::OpenID;					# Package name

use Crypt::OpenID;						# Encryption aspect of OpenID
use strict;							# Strict handling

use Ethereal::Login;						# Login handler
use Ethereal::Web;						# LWP user agent wrapper


################################################################################
# Sub-Routines
################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;					# Self reference
	my $database = shift;					# Database details
	my $param    = shift;					# Parameter handler


	#####################
	# Program Area

	# OpenID Crypto Handle
	my $oid      = new Crypt::OpenID();

	# Create internal hash
	my $self     = {
		AGENT    => undef,				# WWW User Agent
		CGI      => $param->{'XCGI'},			# CGI interface
		CRYPT    => $oid,				# Encryption handler
		DATABASE => $database,				# Database handle,
		PARAM    => $param,				# Parameter handler
		ASSOC    => 'HMAC-SHA1',			# Default association type
		DUMB     => 'dumb',				# Dumb mode
		SMART    => 'smart',				# Smart mode
		SERVER   => 'localhost',			# Default server
		ENCRYPT  => 0,					# Encryption use
		SIGN     => 64					# Signature block size
	 };

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}


####################
# Consumer Check
#
# This is the aspect of Ethereal Realms which will handle the complexities
# associated with consumer handling

sub Consumer_Check
{
	####################
	# Data members
	my $self     = shift;					# Reference to self
	my $ident    = shift;					# Identification to check
	my $server   = shift;					# Server to check against

	my $cgi      = $self->{'CGI'};				# CGI interface
	my $database = $self->{'DATABASE'};			# Database handle
	my $oid      = $self->{'CRYPT'};			# OpenID handler
	my $param    = $self->{'PARAM'};			# Parameter list generated
	my $web;						# User agent encapsulation

	my $mac;						# Encrypted key
	my $pass;						# What to pass on

	my $assoc;						# Association handle
	my $reply;						# Reply
	my $response;						# Server response

	my $inline   = '';					# Inline parameters
	my $sep;						# Seperator
	my $url;						# Url to pass back


	####################
	# Program area

	# Get proper ident, server and associate handle
	($ident, $server, $assoc, $mac) = $self->Discover($ident);

	# Check for existence of server response
	if ((defined($param->{'openid.mode'}))
	  && ($param->{'openid.mode'} =~ /id_res/i))
  	{
		# Acception assertion
		if (defined($param->{'openid.identity'}))
		{
			# What to do next
			# Association failed check against server
			if ((defined($param->{'openid.sig'}))
			 && (defined($param->{'openid.invalidate_handle'})))
		 	{
				# Specific data members
				my %send;
				my %get;

				# Key word
				my $key;


				# Find all appropriate keys
				foreach $key (keys(%{$param}))
				{
					# Transfer over as necessary
					$send{$key} = $param->{$key}
					  if ($key =~ /^openid.*/);
				}

				# Adapt accordingly
				$send{'openid.mode'} = 'check_authentication';


				# Call the server
				$web = new Ethereal::Web($database);

				# With appropriate arguments
				($reply, $response) = $web->Fetch_Post($server, \%send);

				# What is necessary
				if (defined($reply))
				{
					# Figure out what was received
					$self->Parse($reply, \%get);

					# Remove stale handle
					if ((defined($get{'is_valid'}))
					 && (defined($get{'invalidate_handle'}))
					 && ($get{'is_valid'} eq 'true'))
				 	{
						# Remove from database entry
						$database->Quick("DELETE FROM OpenID
							WHERE OpenIDServer=?
							  AND OpenIDAssociate=?",

						  $server,
						  $get{'invalidate_handle'}
						 );

						# Return successful
						return 2;
					}
				}
			}

			# Signature only
			elsif (defined($param->{'openid.sig'}))
			{
				# Return value as necessary
				return 2 if ($self->Signature($mac));
			}
		}

		# Failure is the default
		else
		{
			# We have what we need?
			if (defined($param->{'openid.user_setup_url'}))
			{
				# Print redirection header
				print $cgi->redirect($param->{'openid.user_setup_url'});

				# Return with sufficient notices
				return 1;
			}
		}
	}

	# Initiate connection
	elsif (!defined($param->{'openid.mode'}))
	{
		# Set base params
		$pass  = 'openid.mode'          . '=' . $cgi->escape('checkid_setup') . '&';
		$pass .= 'openid.identity'      . '=' . $cgi->escape($ident) . '&';
		$pass .= 'openid.assoc_handle'  . '=' . $cgi->escape($assoc) . '&';
		$pass .= 'openid.trust_root'    . '=' . $cgi->escape($database->{'SYS'}{'SetOpenIDRoot'}) . '&';

		# Simple Registration Extension
		$pass .= 'openid.sreg.optional' . '=' . $cgi->escape('fullname,email,nickname') . '&';
		$pass .= 'openid.sreg.policy'   . '=' . $cgi->escape($database->{'SYS'}{'SetOpenIDPolicy'}) . '&';


		# Return the process to us
		# Generate self-referencing link
		$url = $cgi->url(-path_info=>1);

		# Generate inlines
		if (defined($param->{'OPENID'}))
		{
			# Encode appropriatly
			$inline = '?SEC=' . $param->EncodeBase64(
				OPENID  => $param->{'OPENID'},
				MORI    => time + 300
			) . '&B64=True';
		}

		# Append accordingly
		$pass .= 'openid.return_to' . '=' . $cgi->escape($url . $inline);

		# Determine seperator
		$sep = ($server =~ /\?/) ? '&' : '?';

		# Redirection line
		print $cgi->redirect($server . $sep . $pass);

		# Successful return
		return 1;
	}

	# Failed
	return 0;
}


####################
# Discover
#
# This part is required to discover if the OpenID component provided is in
# valid and has all of the necessary information to proceede.  Part of this
# process will also be to associate with the OpenID server and pull the shared
# key.

sub Discover
{
	####################
	# Data members
	my $self     = shift;					# Reference to self
	my $ident    = shift;					# OpenID address

	my $assoc;						# Associate handle
	my $mac;						# Mac Key
	my $reply;						# Reply from their server
	my $response;						# Results in server address
	my $server;						# Server address

	my $state    = 'smart';					# Default state for us

	my $oid      = $self->{'CRYPT'};			# OpenID crypto
	my $database = $self->{'DATABASE'};			# Database handle
	my $web;						# User agent encapsulation

	my %send;						# Arguments to pass
	my %get;						# Arguments pulled

	####################
	# Program area

	#
	# Basic fact checking

	# Return if an IP address
	return undef, undef, undef, undef if ($ident =~ /\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}/);

	# Correct any missing elements from the HTML
	$ident = 'http://' . $ident unless ($ident =~ /^http[s]?:\/\//i);


	# Pull the page
	# Create instance
	$web = new Ethereal::Web($database);

	# Try and see
	($reply, $response) = $web->Fetch_Get($ident);

	# Fail if something fails
	return undef, undef, undef, undef, $response unless (defined($reply));


	# Start to find information
	# First necessary component is found
	if ($reply =~ /<link.*rel="openid.server".*href="(.*)?".*>/i)
	{
		# Assign appropriatly
		$server = $1;

		# Is there a delegate?
		$ident  = $1
		  if ($reply =~ /<link.*rel="openid.delegate".*href="(.*)?".*>/i);
	}


	# Associate handle verification
	# Check for existing one
	($assoc, $mac) = $database->DataGet("SELECT
		  OpenIDAssociate AS \"OpenIDAssociate\",
		  OpenIDMac       AS \"OpenIDMac\"
		FROM OpenID
		WHERE OpenIDServer=?
		  AND OpenIDType=?
		  AND OpenIDExpires > LOCALTIMESTAMP
		  AND OpenIDState=?",

	  $server,
	  $self->{'ASSOC'},
	  'smart'
	);


	# Need to associate once again
	unless (defined($assoc))
	{
		# Time to contact the server
		# And associate to get a shared secret
		%send = (
			'openid.mode'       => 'associate',
			'openid.assoc_type' => $self->{'ASSOC'}
		);

		# Is encryption an option
		if ($self->{'ENCRYPT'})
		{
			# Generate a key
			$oid->generate_keys(
				Size     => 64,
				Verbose  => 0 
			);

			# Add the additional components
			$send{'openid.session_type'}       = 'DH-SHA1';
			$send{'openid.dh_modulus'}         = $oid->b64_btwoc($oid->{'p'});
			$send{'openid.dh_gen'}             = $oid->b64_btwoc($oid->{'g'});
			$send{'openid.dh_consumer_public'} = $oid->b64_btwoc($oid->{'pub_key'});
		}

		# Call the server
		# With appropriate arguments
		($reply, $response) = $web->Fetch_Post($server, \%send);

		# Fail if something fails
		return undef, undef, undef, undef, $response unless (defined($reply));

		# Figure out what was received
		$self->Parse($reply, \%get);


		# Extraction necessary?
		if ((defined($get{'enc_mac_key'}))
		 && (defined($get{'dh_server_public'})))
	 	{
			# Extract encoded key
			$get{'mac_key'} = $oid->extract_shared(\%get)
		}

		# Something went wrong
		return undef, undef, undef, undef, undef
		  unless (defined($get{'mac_key'}));


		# Write and verify
		#
		eval
		{
			# Interval provided with five less minutes to prevent
			# use of expired keys
			$database->Quick("INSERT INTO OpenID
				(OpenIDServer,
				 OpenIDType,
				 OpenIDAssociate,
				 OpenIDMac,
				 OpenIDExpires,
				 OpenIDState)
				VALUES (?,?,?,?,addInterval('$get{'expires_in'} minutes'),?)",

			  $server,
			  $send{'openid.assoc_type'},
			  $get{'assoc_handle'},
			  $get{'mac_key'},
			  $self->{'SMART'}
			 );
		};

		# Failed
		if ($@)
		{
			# Display a warning
			warn("There was an error in attempting to save OpenID state", $@);

			# Return without success
			return undef, undef, undef, undef;
		}


		# Set accordingly
		$assoc = $get{'assoc_handle'};
		$mac   = $get{'mac_key'};
	}
		
	# Return everything
	return $ident, $server, $assoc, $mac;
}


####################
# Keys
#
# Simple but ads consistency required to generate the AssociateID and MacID
# keys used in transactions.

sub Keys
{
	####################
	# Data members
	my $self     = shift;					# Reference to self

	my $oid      = $self->{'CRYPT'};			# OpenID crypto handler
	my $param    = $self->{'PARAM'};			# Parameters
	my $database = $self->{'DATABASE'};			# Database handle

	my $assoc;						# Associate key
	my $secret;						# Generated Mac ID

	####################
	# Program area

	# Generate associate identifier
	$assoc  = $database->{'SEARCH'} . ':' . time . ':' . $self->Random(15);

	# Generate mac key
	$secret = $oid->make_mac($assoc, $database->{'SYS'}{'SetKey'});

	# Since the size of the mac key cannot exceed the size of the hash
	# it becomes easier to just pass it through the hash to make sure 

	# Return both
	return $assoc, $secret;
}


####################
# Parse
#
# Will parse the output of a returned POST file necessary to get extract the
# information within

sub Parse
{
	####################
	# Data members
	my $self   = shift;					# Reference to self
	my $reply  = shift;					# Reply to parse
	my $args   = shift;					# Arguments hash ref

	my $key;						# Key name
	my $val;						# Key value


	####################
	# Program area

	# Figure out what was received
	while ($reply =~ /^(.+?)\:(.+)$/gm)
	{
		# Precautions
		$key = $1;
		$val = $2;

		# Add to the hash
		$args->{$key} = $val;
	}
}


####################
# Provider
#
# The server side of the house.  Will take in specific output and based on the data will
# eventually authenticate the user.

sub Provider
{
	####################
	# Data members
	my $self     = shift;					# Reference to self

	my $cgi      = $self->{'CGI'};				# CGI interface
	my $database = $self->{'DATABASE'};			# Database handle
	my $oid      = $self->{'CRYPT'};			# OpenID handler
	my $param    = $self->{'PARAM'};			# Parameter list generated

	my $assoc;						# Associate ID
	my $key;						# Individual key
	my $mode;						# Modis operandi
	my $secret;						# Mac key

	my %login;						# Authentication hash
	my %keys;						# Keys to display


	####################
	# Program Area

	# Verification and associations are restricted to this
	# mode.   Any attempts at these in GET are to be ignored
	# for safety reasons.
	if ((defined($param->{'openid.mode'}))
	 && ($param->{'openid.mode'} eq 'associate')
	 && ($cgi->request_method() eq 'POST'))
 	{
		# Generate necessary elements
		($assoc, $secret) = $self->Keys();

		# Initialize response
		%keys = (
			assoc_type   => 'HMAC-SHA1',
			assoc_handle => $assoc,
			expires_in   => $database->{'SYS'}{'SetOpenIDSmart'},
			mac_key      => $secret
		 );


		#
		# Encryption enabled
		if ($self->{'ENCRYPT'})
		{
			# These are ignored otherwise
			if ((defined($param->{'openid.session_type'}))
			 && (defined($param->{'openid.dh_modulus'}))
			 && (defined($param->{'openid.dh_gen'}))
			 && (defined($param->{'openid.dh_consumer_public'}))
			 && ($param->{'openid.session_type'} eq 'DH-SHA1'))
		 	{
				# Encrypt the package
				$keys{'enc_mac_key'} = $oid->make_shared($param, $oid->d64($secret));

				# Establish public key
				$keys{'dh_server_public'} = $oid->b64_btwoc($oid->{'pub_key'});

				# Establish session type
				$keys{'session_type'} = 'DH-SHA1';
			}
		}

		# Save association to the database
		$database->Quick("INSERT INTO OpenID
			(OpenIDServer,
			 OpenIDType,
			 OpenIDAssociate,
			 OpenIDMac,
			 OpenIDExpires,
			 OpenIDState)
			VALUES (?,?,?,?,addInterval('$keys{expires_in} minutes'),?)",

		  $self->{'SERVER'},
		  $keys{'assoc_type'},
		  $keys{'assoc_handle'},
		  $keys{'mac_key'},
		  $self->{'SMART'}
		 );

		# Cleanup
		delete($keys{'mac_key'})
		 if defined($keys{'enc_mac_key'});


		# Display to screen
		print $cgi->header();

		# Loop and display keys
		foreach $key (keys(%keys))
		{
			# Display as per spec
			print $key . ":" . $keys{$key} . "\n";
		}

		# Everything is succesful
		return 1;
	}

	# Check authentication has the same limitations as
	# above.  Any and all attempts at using methods other
	# then post.
	elsif ((defined($param->{'openid.mode'}))
	 && ($param->{'openid.mode'} eq 'check_authentication')
	 && ($cgi->request_method() eq 'POST'))
	{
		# Local
		my $result;				# Result generated

		# What is necessary is there
		if ((defined($param->{'openid.sig'}))
		 && (defined($param->{'openid.signed'}))
		 && (defined($param->{'openid.assoc_handle'})))
	 	{
			# Made sure the Associate ID is valid
			($secret) = $database->DataGet("SELECT
				  OpenIDMac
				FROM OpenID
				WHERE OpenIDServer=?
				  AND OpenIDAssociate=?
				  AND OpenIDState=?",
				

			  $self->{'SERVER'},
			  $param->{'openid.assoc_handle'},
			  $self->{'DUMB'}
			 );

			# Only confirm in dumb mode
			if (defined($secret))
		 	{
				# Establish mode
				$param->{'openid.mode'} = 'id_res';


				# Signature passed
				$result = ($self->Signature($secret))
				 ? 'true'
				 : 'false';


				# Display to the screen
				print $cgi->header();

				# Basic elements
				print "openid.mode:id_res\n";
				print "is_valid:" . $result . "\n";

				# Only if necessary
				print "invalidate_handle:" . $param->{'openid.invalidate_handle'} . "\n"
				  if (defined($param->{'openid.invalidate_handle'}));

				# Good result either way
				return 1;
			}
		}
	}

	# This mode will do the actual authentication
	elsif ((defined($param->{'openid.mode'}))
	 && (($param->{'openid.mode'} eq 'checkid_immediate')
 	  || ($param->{'openid.mode'} eq 'checkid_setup')))
	{
		# Safety check
		#
		# Make sure all the necessary elements are in fact there
		# and make sure the root is a component of the return_to
		# address.
		if ((defined($param->{'openid.identity'}))
		 && (defined($param->{'openid.return_to'}))
		 && (defined($param->{'openid.trust_root'}))
		 && ($param->{'openid.return_to'} =~ /^$param->{'openid.trust_root'}/i))
	 	{
			# Login handle
			my $login = new Ethereal::Login($database, $cgi, $param);

			# Link with hash
			$database->GetHashLogin(\%login);


			# Disable header generation
			$login->{'AUTOMATE'} = 0;

			# Signature block base
			my @sig = qw(mode identity return_to);  # Signature block


			# Prevent screwups
			$param->{'USER_SECURE'} = 'On';


			# See if the user has canceled the step
			# Return to consumer
			if (defined($param->{$login{'TxtOpenIDCancel'}}))
			{
				# Just do the redirect
				print $self->Redirect_Cancel();
			}

			# See if the user can be authenticated
			# remove the ability to use OpenID as an
			# identifier.
			elsif (($login->GetVerificationOpenID($self))
			 && ($database->{'SYS'}{'SetOpenIDIdentity'} =~ /^(.*?)MNAME(.*)$/))
			{
				# Local
				my $elem;				# Singular element
				my $pass;				# Element to pass on

				my $uid;				# UserID to compare
				my $pre   = $1;				# Pre MNAME section
				my $post  = $2;				# Post MNANE section



				# Massage accordingly for consistency
				$uid = (($database->{'SYS'}{'SetOpenIDTrailing'} =~ /true/i)
			       	     && ($param->{'openid.identity'} !~ /^.*\/$/))
			  	  ? $param->{'openid.identity'} . '/'
				  : $param->{'openid.identity'};
				$uid =~ s/^$pre(.*?)$post$/$1/;

				# What is the default handle
				($uid) = $database->DataGetPuppetShort($uid);


				# Match usernames for both
				if ($param->{'USER'} eq $uid)
				{
					# Authenticated user and login both match
					# Now check for trust issues
					if ($self->Trust(\%login))
					{
						# Setup keys
						%keys = (
							'openid.mode'         => 'id_res',
							'openid.identity'     => $param->{'openid.identity'},
							'openid.return_to'    => $param->{'openid.return_to'}
						);


						# Check for Associate ID
						# But only if required
						($assoc, $secret) = $database->DataGet("SELECT
							  OpenIDAssociate,
							  OpenIDMac
							FROM OpenID
							WHERE OpenIDServer=?
							  AND OpenIDAssociate=?
							  AND OpenIDState=?",

						  $self->{'SERVER'},
						  $param->{'openid.assoc_handle'},
						  $self->{'SMART'}
						) if (defined($param->{'openid.assoc_handle'}));

						# Handling of associate id
						unless (defined($assoc))
						{
							# Invalid handle list
							$keys{'openid.invalidate_handle'} = $param->{'openid.assoc_handle'}
							  if (defined($param->{'openid.assoc_handle'}));

							#
							# Generate necessary elements
							($assoc, $secret) = $self->Keys();

							# Save association to the database
							$database->Quick("INSERT INTO OpenID
								(OpenIDServer,
								 OpenIDType,
								 OpenIDAssociate,
								 OpenIDMac,
								 OpenIDExpires,
								 OpenIDState)
								VALUES (?,?,?,?,addInterval('$database->{SYS}{SetOpenIDDumb} minutes'),?)",

							  $self->{'SERVER'},
							  $self->{'ASSOC'},
							  $assoc,
							  $secret,
							  $self->{'DUMB'}
							 );
						}

						# Handle components
						$keys{'openid.assoc_handle'} = $assoc;


						# Additional registration components
						if ((defined($param->{'openid.sreg.required'}))
						 || (defined($param->{'openid.sreg.optional'})))
					 	{
							my $demand = '';	# Registration additions
							my $res;		# Results list
							my $statement;		# Query statement

							my @sregs;		# Registration elements


							# What we got
							$database->Pull(\$statement, "SELECT
								u.PuppeteerName  AS \"openid.sreg.fullname\",
								u.PuppeteerEmail AS \"openid.sreg.email\",
								h.PuppetName     AS \"openid.sreg.nickname\"
								FROM Puppeteer u, Puppet h
								WHERE u.PuppeteerLogin = h.PuppeteerLogin
								  AND u.PuppeteerLogin=?
								  AND h.PuppetDefault=?",

							  $param->{'USER'},
							  'yes');

						  	# Query worked
						  	if (($res) = $statement->fetchrow_hashref())
							{
						 	 	#
								# What do they want
								# Required fields
								$demand .= $param->{'openid.sreg.required'}
								 if (defined($param->{'openid.sreg.required'}));

								# Not required
								$demand .= $param->{'openid.sreg.optional'}
								 if (defined($param->{'openid.sreg.optional'}));

								# Split their demans
								@sregs = split(/,/, $demand);

								# Loop find and add
								foreach $elem (@sregs)
								{
									# We have a match, so transfer
									if (defined($res->{'openid.sreg.' . $elem}))
									{
										# Add if necessary
										$keys{'openid.sreg.' . $elem} = $res->{'openid.sreg.' . $elem};

										# Push onto signature array
										push(@sig, 'sreg.' . $elem);
									}
								}
							}

							# Cleanup
							$statement->finish();
						}


						# Signature
						# Elements to be signed
						$keys{'openid.signed'} = join(',', @sig);


						# Display redirection
						print $self->Redirect(\%keys, \@sig, $secret);

						# Successful
						return 1;
					}

					#
					# User no longer trusts
				}

				#
				# Not the same user
				else
				{
					# Redirect user out of here
					print $self->Redirect_Cancel();
				}
			}

			#
			# Login failed handle appropriatly
			return 0;
		}
	}


	# Display error
	print $cgi->header();
	print $database->{'SYS'}{'SetOpenIDError'} . "\n";


	# Fail miserably
	return 0;
}


####################
# Random
#
# Generates a random string from a pool of characters pre-determined during
# the setup.  Nothing complex or necessarily 100% secure.

sub Random
{
	####################
	# Data members
	my $self   = shift;					# Reference to self
	my $length = shift;					# Length to generate

	my $digits;						# Pool to draw from
	my $string;						# Generated string


	####################
	# Program area

	# Safety
	$length = 24
	  unless ((defined($length)) && (length($length) < 256));

	# Initialize pool
	$string = '';
	$digits = 'abcdefghijklmnopqrstuvwzyzABCDEFGHIJKLMNOPQRSTUVWZYZ0123456789';

	# Loop through
	for (1..$length)
	{
		# Append accordingly
		$string .= substr($digits, int(rand(62)), 1);
	}

	# Return string
	return $string;
}


####################
# Redirect
#
# Handles the basics of redirection when required for the identifier provider
# part of the spec.

sub Redirect
{
	####################
	# Data members
	my $self   = shift;					# Reference to self
	my $keys   = shift;					# Keys to sign
	my $sig    = shift;					# Signature items
	my $secret = shift;					# Secret mac key

	my $cgi    = $self->{'CGI'};				# Common gateway interface
	my $param  = $self->{'PARAM'};				# Parameters
	my $oid    = $self->{'CRYPT'};				# OpenID Encryption

	my $elem;						# Singular element
	my $pass;						# Passage of information


	####################
	# Program area

	#
	# Signature specific
	if (defined($sig))
	{
		# Define
		$pass = '';

		# Seek out and add in
		foreach $elem (@{$sig})
		{
			# Append accordingly
			$pass .= $elem . ':' . $keys->{'openid.' . $elem} . "\n";
		}

		# Get signature
		$keys->{'openid.sig'} = $oid->make_sig($pass, $secret, $self->{'SIGN'});
	}

	#
	# Establish passing line
	$pass = '';

	# Combine into query line
	foreach $elem (sort(keys(%{$keys})))
	{
		# Append and escape
		$pass .= '&' . $elem . '=' . $cgi->escape($keys->{$elem});
	}


	# Prepare for our use
	# By removing any escape characters
	$param->{'openid.return_to'} = $cgi->unescape($param->{'openid.return_to'});


	# Is the first character wrong?
	# Because other form elements exist
	$pass = ($param->{'openid.return_to'} =~ /\?/)
	  ? $pass
	  : '?' . substr($pass, 1);

	# Return completed line
	return $cgi->redirect($param->{'openid.return_to'} . $pass);
}

# Alias for simple cancel
sub Redirect_Cancel
{ return Redirect(shift, {'openid.mode' => 'cancel'}, @_) };


####################
# Signature
#
# Handles signatures considering that the operation is the same for both sides
# of the server.

sub Signature
{

	####################
	# Data members
	my $self   = shift;					# Reference to self
	my $secret = shift;					# Secret mac key

	my $oid    = $self->{'CRYPT'};				# OpenID crypto handler
	my $param  = $self->{'PARAM'};				# Parameters

	my $elem;						# Singular element
	my $pass;						# Line to pass and check
	my $sig;						# Final signature

	my @sig;						# Signature elements


	####################
	# Program area

	# Initialize
	$pass = '';

	# Safety
	$param->{'openid.signed'} = 'mode,identity,return_to'
	  unless (defined($param->{'openid.signed'}));


	# Form the line
	# Check for signature
	@sig = split(/,/, $param->{'openid.signed'});

	# Seek out and add in
	foreach $elem (@sig)
	{
		# Append accordingly
		$pass .= $elem . ':' . $param->{'openid.' . $elem} . "\n";
	}


	# Get signature
	$sig = $oid->make_sig($pass, $secret, $self->{'SIGN'});


	# Return value as necessary
	return 1 if ($sig eq $param->{'openid.sig'});

	# Default to failure otherwise
	return 0;
}


#####################
# Trust
#
# Part of the OpenID specification in which the user has to trust the server
# which is requesting to authenticate them.   This is not part of the normal
# authentication process because it is not intended to be used that often.

sub Trust
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $login    = shift;					# Login hash

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'DATABASE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $header;						# Header to display
	my $tmpl;						# Template handler
	my $menu;						# Menu handler

	my $delay;						# Delay time
	my $regist   = '';					# Registration leech
	my $test;						# Site to trust 
	my $trust;						# Trust issues
	
	my %keys;						# Keys to pass on


	#####################
	# Program area
	# OpenID redirection

	# What to check up on trust
	$test = (defined($param->{'openid.trust_root'}))
	  ? $param->{'openid.trust_root'}
	  : $param->{'openid.return_to'};


	# Management of trust
	if (defined($param->{'TRUST'}))
	{
		# Establish trust
		$trust = ((defined($param->{$login->{'TxtOpenIDDay'}}))
		       || (defined($param->{$login->{'TxtOpenIDAlways'}})))
		  ? 'trust'
		  : 'never';

		# Type of delay
		$delay = (defined($param->{$login->{'TxtOpenIDDay'}}))
		  ? "addInterval(\'1 day\')"
		  : "NULL";


		# Update database
		$database->Quick("INSERT INTO PuppeteerTrust
			(PuppeteerLogin,
			 OpenIDRoot,
			 OpenIDExpires,
			 OpenIDTrust)
			VALUES (?,?,$delay,?)",

		  $param->{'USER'},
		  $test,
		  $trust
		 );

		# Cleanup
		$param->Cleanup('TRUST');
	}

	# Do we trust them
	unless (defined($trust))
	{
		# Check up on trust issues now
		($trust) = $database->DataGet("SELECT OpenIDTrust
			FROM PuppeteerTrust
			WHERE PuppeteerLogin=?
			  AND OpenIDRoot=?
			  AND (OpenIDExpires IS NULL
			    OR OpenIDExpires > LOCALTIMESTAMP)",
	
		 $param->{'USER'},
		 $test
		);
	}

	# Return successful right now if valid
	if ((defined($trust))
	 && ($trust eq 'trust'))
 	{
		# Successful and return as such
		return 1;
	}

	# We do not trust this individual
	elsif ((defined($trust))
	 && ($trust eq 'never'))
 	{
		# Display redirection
		print $self->Redirect_Cancel();

		# Something went wrong
		return 0;
	}


	# Impatient mode
	# Redirect to client and let them deal with it
 	if (($param->{'openid.mode'} eq 'checkid_immediate')
	 && (!defined($param->{'BYPASS'})))
 	{
		# Initial setting
		$keys{'openid.mode'} = 'id_res';

		# Establish return url
		$keys{'openid.user_setup_url'}
		  = $cgi->url(-path_info=>1) 
		  . $param->Crypt($param->OpenID_Flat(),
			  USER   => $param->{'USER'},
			  CRYPT  => $param->{'CRYPT'}, 
			  BYPASS => 'True');

		# Display redirect
		print $self->Redirect(\%keys);

		# Return like the other
		return 0;
	}

	# Cleanup
	$param->Cleanup('BYPASS');
	

	# Need to gain trust
	# Display information only if necessary

	# Create instances
	$tmpl     = new Ethereal::Template($login);
	$menu     = new Ethereal::Menu($database, $param);

	# Registration
	$regist = $tmpl->Pass('TmplRegister', MREG=>$param->{'openid.sreg.policy_url'})
	  if (defined($param->{'openid.sreg.policy_url'}));



	# HTML
	# Setup title
	$database->{'TITLE'} = $database->{'SYS'}{'TitTrust'};

	# Extract header
	$header = $database->DocumentGetHeader();

	# Headers
	print $cgi->header();
	print $menu->Rewrite($header), "\n";


	# Information display
	# Start from and embed
	print $cgi->start_form();
	print $param->EmbedNormal($param->OpenID_Flat(),
		USER  => $param->{'USER'},
		CRYPT => $param->{'CRYPT'},
		TRUST => 'True'), "\n";

	# Display template
	$tmpl->Show('TmplTrust',
		MTITLE => $database->{'SYS'}{'TitTrust'},
		MROOT  => $test,
		MREGS  => $regist,
		WTDAY  => $cgi->submit($login->{'TxtOpenIDDay'}),
		WTALW  => $cgi->submit($login->{'TxtOpenIDAlways'}),
		WTNVR  => $cgi->submit($login->{'TxtOpenIDNever'})
	 );

	# End of form
	print $cgi->end_form(), "\n";

	# Display footer
	print $database->DocumentGetFooter();

	# Fails by default
	return 0;
}


1;								# Return true
