################################################################################
# Created       : Martin Foster
# Modified      : 14-Dec-2006
################################################################################
#
# Words - Component of Ethereal Realms handling word seperation, weight and
#         stop words listing. 
# Copyright (C) 2000-2006  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 3-3456 Wolfe Cres
#                 Halifax, Nova Scotia
#                 B3L 3S2
#
#################################################################################

package Ethereal::Words;					# Package name

use Lingua::Stem;						# Language stemmer
use Lingua::EN::Tagger;						# Part-of-speech tagger for English
								# natural language processing
use Math::Trig;


################################################################################
# Sub-Routines
################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data Members
	my $class    = shift;					# Self reference
	my $stem;						# Stemming language

	#####################
	# Program Area

	# Create instance
	$stem = Lingua::Stem->new(-locale => 'EN-UK');
    	$stem->stem_caching({ -level => 2 });
	

	# Create internal hash
	my $self     = {
		NORMAL => 10,					# Normalization max
		STEM   => $stem,				# Stemming handle
		STOP   => undef					# Stop words
 	 };

	# Bless $self into designated class
	bless $self, $class;

	# Return self
	return $self;
}

#####################
# Clean
#
# Will clean up the hash for use by a search engine by using Lingua::Stem
# to find the root of a word.    Before this happens the stop words will
# be applied

sub Clean
{
	#####################
	# Data members

	my $self     = shift;					# Self referencing link

	my $stop;						# Stop words hash
	my $root;						# Find root
	my $word;						# Singular word being worked on

	my %words    = @_;
	my %clean;						# New cleaner hash


	#####################
	# Program area

	# Generate stopwords
	$self->Stopword() unless (defined($self->{'STOP'}));

	# Loop through
	foreach $word (sort(keys(%words)))
	{
		# Stop word reached
		next if (defined($self->{'STOP'}->{$word}));

		# Find root 
		# If it is not a word phrase
		$root = $self->Stem($word) unless ($word =~ /\s/);

		# Add to a new hash
		$clean{$root} = (defined($clean{$root}))
		 ? $clean{$root} + $words{$word}
		 : $words{$word}; 
	}

	# Return value
	return %clean;
}


####################
# Query
#
# Query builder, will determin the action to take depending on
# what was recieved in the parameter list. 

sub Query
{
	####################
	# Data members
	my $self   = shift;				# Reference to self
	my $string = shift;				# String to figure out

	my $format;					# Formatted options
	my $word;					# Current word

	my %words;					# Array of dismembered words


	####################
	# Program area

	# Basics
	# Split words
	%words = $self->Split($string);

	# Clean it up
	%words = $self->Clean(%words);


	# Formatting
	# Initialize
	$format = '';
	
	# Looping
	foreach $word (keys(%words))
	{
		# Implement
		$format .= "\'" . $word . "\',";
	}

	# Remove trailing comma
	chop($format);


	# Return
	return $format;
}


#####################
# Reverse 
#
# Reverse index will automate the entire process of getting unique words
# and scores for word phrases.

sub Reverse
{
	#####################
	# Data members

	my $self     = shift;					# Self referencing link
	my $database = shift;					# Database handle
	my $gid      = shift;					# Gallery ID
	my $wid      = shift;					# Workid ID

	my $field;						# Current field
	my $multi;						# Multiplier

	my @fields;						# Fields to loop through
	my @multi    = qw(1 0.5 0.75);				# Multipliers

	my %curr;						# Current hash
	my %full;						# Complete hash


	#####################
	# Program area

	# Create fields - Title, Description and Text
	push(@fields, shift, shift, shift);	

	# Pull scores
	foreach $field (@fields)
	{
		# Mutiplier
		$multi = shift(@multi);

		# Reset
		%curr = ();

		# Safety
		if (defined($field))
		{
			# Split into unique words
			%curr  = $self->Split($field);

			# Clean up and stem
			%curr  = $self->Clean(%curr);

			# Add scores
			$self->Weight(\%curr, $multi);
		}

		# Append to next hash
		foreach $field (keys(%curr))
		{
			# Append or create
			$comp{$field} = (defined($comp{$field})) 
			  ? $comp{$field} + $curr{$field}
			  : $curr{$field};
		}
	}

	# Database handling
	# Clear out field
	$database->Write("DELETE FROM Index
			WHERE GalleryID=? AND WorkID=?", $gid, $wid);

	# Append to the database
	foreach $field (keys(%comp))
	{
		# Insert into field
		$database->Write("INSERT INTO Index
			(GalleryID,
			 WorkID,
			 IndexWord,
			 IndexScore) VALUES (?,?,?,?)",

		 $gid,
		 $wid,
		 $field,
		 $comp{$field}
		);
	}

	# Reset
	%curr = ();
	%comp = ();

	# Return true
	return 1;
}


#####################
# Stopword gen
#
# Stopword generation sub-routine.   Essentially takes in a full list of stop
# words used to reduce false positives in a search engine.

sub Stopword
{
	#####################
	# Data members

	# Self referencing link
	my $self     = shift;

	# Singular word list
	my $word;
	
	# Generate rather large list
        my @words = qw(about above according across actually adj after
afterwards again against all almost alone along already also although
always among amongst an and another any anyhow anyone anything
anywhere are arent around as at be became because become becomes
becoming been before beforehand begin beginning behind being below
beside besides between beyond billion both but by can can't cannot
caption co could couldnt did didn't do does doesnt dont down during
each eg eight eighty either else elsewhere end ending enough etc even
ever every everyone everything everywhere except few fifty first five
for former formerly forty found four from further had has hasn't have
haven't he hed hell hes hence her here heres hereafter hereby herein
hereupon hers herself him himself his how however hundred i'd i'll
i'm i've ie if in inc indeed instead into is isn't it its it's itself
know last later latter latterly least less let lets like likely ltd
made make makes many maybe me meantime meanwhile might million miss
more moreover most mostly mr mrs much must my myself namely neither
never nevertheless next nine ninety no nobody none nonetheless noone
nor not nothing now nowhere of off often on once one ones only onto
or other others otherwise our ours ourselves out over overall own per
perhaps rather recent recently same seem seemed seeming seems seven
seventy several she shed she'll shes she's should shouldnt since six
sixty so some somehow someone something sometime sometimes somewhere
still stop such taking ten than that that'll thats that've the their
them themselves then thence there thered therell therere theres
thereve thereafter thereby therefore therein thereupon these they
theyd they'll theyre they've thirty this those though thousand three
through throughout thru thus to together too toward towards trillion
twenty two under unless unlike unlikely until up upon us used using
very via was wasn't we we'd well were weve were werent what what'll
whats what've whatever when whence whenever where wheres whereafter
whereas whereby wherein whereupon wherever whether which while
whither who whod wholl whos whoever whole whom whomever whose why will
with within without won't would wouldn't yes yet you you'd you'll 
you're you've your yours yourself yourselves);


	#####################
	# Program area	

	# Initialize
	$self->{'STOP'} = {};

	# Populate hash
	while ($word = shift(@words))
	{
		# Create an empty hash
		$self->{'STOP'}->{$word} = 1;
	}

	# Basic return
	return 1;
}


#####################
# Word split
#
# Generates the necessary words used for spell checking and search
# search engines.  By default all words are forced to lower case.

sub Split
{
	#####################
	# Data members

	my $self     = shift;					# Self referencing link
	my $str      = shift;					# String to manage
	my $choice   = shift;					# Choice of the matter

	my $tagger;						# Part of speed tagger

	my $word;						# Singular word

	my @words;						# Word list
	my %words;						# Words list w/ score


	#####################
	# Program area

	# Safety
	$choice = (defined($choice)) ? $choice : 0;


	# Tags and special characters
	$str =~ s/<[^>]*>//gs;
	$str =~ s/\&.*?\;//gs;

	# Comments
	$str =~ s/<!-|-->//gs;


	# Old style
	# Each word as is
	if ($choice != 1)
	{
		# Filter string for all words
		while ($str =~ /([a-zA-Z0-9\xc0-\xff\_\-\'\.]{2,})/gs)
		{
			# Assign word
			$word =  lc($1);

			# Remove leading and trailing punctuation
			$word =~ s/^[^a-zA-Z0-9\xc0-\xff\_\-]+//s;
			$word =~ s/[^a-zA-Z0-9\xc0-\xff\_\-]+$//s;
			

			# Keep track of wording
			$words{$word} = (defined($words{$word})) 
			 ? $words{$word} + 1
			 : 1;
		}
	}

	# Part-of-speech tagger
	else
	{
		# Create instance
		# With stemming in place
		$tagger = Lingua::EN::Tagger->new(
			stem  => 1,				# Stem using Lingua::Stem::EN
			relax => 1				# Relax the Hidden Markov Model
		 );

		# Get unique words
		%words = $tagger->get_words($str);
	}

	# Return words
	return %words;
}

# Splits with HTML and Taggert
sub SplitWTaggert
{ return Split(shift, shift, 1); }


#####################
# Stemmer
#
# Used to hide the rather grotesque format in which Lingua::Stem actually
# uses to work.  Taken from 'Building a Vector Space Search Engine in Perl'
# by Maciej Ceglowski on Perl.com.

sub Stem 
{
	#####################
	# Data members

	my $self     = shift;					# Self referencing link
	my $word     = shift;					# Word to stem

	my $stem     = $self->{'STEM'};				# Stemming handle
	my $ref;						# Stemmer results reference


	#####################
	# Program area

	# Pull stem reference
	$ref = $stem->stem($word);

	# Return accordingly
	return $ref->[0];
}


#####################
# Weight
#
# Calculates the weight that a word will have based on the search
# criteria met.

sub Weight
{
	#####################
	# Data members

	my $self     = shift;					# Self referencing link
	my $words    = shift;					# Hash of words to be weighed
	my $multiply = shift;					# Multiplier used in certain cases

	my $max      = $self->{'NORMAL'};			# Maximum size to look for
	my $word;						# Singular word being worked on


	#####################
	# Program area

	# Safety
	$multiply = (defined($multiply)) ? $multiply : 1;

	# Looping
	foreach $word (sort(keys(%{$words})))
	{
		# Apply multiplier
		$words->{$word} = $words->{$word} * $multiply;

		# Normalize
		$words->{$word} = int(sin(($words->{$word}/$max)/(3.14/4)) * 1000);
	}	
}

1;								# Return true
